/* ***********************************************************
 * Copyright (c) 2006, 2008 IBM Corporation and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * $Id: DefaultChartRenderXYSurface.java,v 1.4 2008-05-23 14:11:50 jcayne Exp $
 *
 * Contributors:
 * IBM - Initial API and implementation
 ************************************************************/

package org.eclipse.tptp.platform.report.chart.internal;

import org.eclipse.tptp.platform.report.core.internal.*;
import org.eclipse.tptp.platform.report.drawutil.internal.IGCDStyle;
import org.eclipse.tptp.platform.report.drawutil.internal.Point2D;
import org.eclipse.tptp.platform.report.drawutil.internal.Point3D;
import org.eclipse.tptp.platform.report.drawutil.internal.Vector3D;
import org.eclipse.tptp.platform.report.igc.internal.IGC;
import org.eclipse.tptp.platform.report.igc.internal.IPen;
import org.eclipse.tptp.platform.report.igc.internal.IRect;
import org.eclipse.tptp.platform.report.igc.internal.IShape;
import org.eclipse.tptp.platform.report.igc.util.internal.GradientGenerator;
import org.eclipse.tptp.platform.report.igc.util.internal.LineStylePen;
import org.eclipse.tptp.platform.report.igc.util.internal.Polygon;
import org.eclipse.tptp.platform.report.igc.util.internal.RGBA;
import org.eclipse.tptp.platform.report.igc.util.internal.Radian;
import org.eclipse.tptp.platform.report.igc.util.internal.Rect;
import org.eclipse.tptp.platform.report.igc.util.internal.SolidBrush;
import org.eclipse.tptp.platform.report.tools.internal.IDisposable;
import org.eclipse.tptp.platform.report.tools.internal.VDouble;


/**
 * Default Chart Render for  T_XY_SURFACE graphic type.
 * @see DGraphic
 * @deprecated As of TPTP 4.5.0, use the TPTP Business Intelligence and Reporting Tools (BIRT) reporting infrastructure (<code>org.eclipse.tptp.platform.report.birt</code>).
 * 
 */
public class DefaultChartRenderXYSurface
{
  
  /** variant part of pertistent data for this graphic */
  static class H3D implements IDisposable
  {
    DAxis z_axis, x_axis, y_axis; //optional
    IGCDStyle z_axis_style, x_axis_style, y_axis_style;
    
    Box3D bbox;
    float shadow; //P_3D_SHADOW
    float light; //P_3D_LIGHT
    IXYSurface surface ;
    IDItem     loc_item; //item used for locatoin
    
    
    double contour_line; //from property, -1 for none
    boolean contour_line_z_axis; //true if delta is defined for z axis
    int    contour_line_rgba;// color
    IXYSurfaceColorizer contour_line_colorizer; //null if constant color
    boolean projected_contour_line; //if P_XYZ_ZPROJ is set with contour_line
    double  zproj_contour_line;
    
    boolean draw_polygon; //inverse of P_XYS_NO_POLYGON property
    
    public void dispose()
    {
    }
  }
  
  private static boolean isValidPersistData( H3D p, DefaultChartRenderData d, RenderPersistData rpd )
  {
    if( p==null ) return false;
    return true;
  }

  /** 
   * Render surface 
   * Rendering is made AsIs, no Z-Buffer, no hidden face remove algorythm, here is a simple
   * 3D drawing.
   */
  public static void render( DefaultChartRenderData d, RenderPersistData rpd )
     throws DefaultRenderChartLocation
  {
    H3D p = null;   
    if( rpd.g instanceof H3D ) p =(H3D)rpd.g;

    //rebuild data ?
    if( !isValidPersistData(p, d, rpd) )
    {
      p = rebuildPersistData( d, rpd );
    }
    
    if( p.surface==null )
    {
      if(d.drawing()) 
      {
        d.error( "Error: missing IXYSurface as child of DGraphic", rpd );
      }
      return ;
    }
    //note: Z axis is not mandatory but used for render only
    
    
    {
      DPropertyStore store = rpd.graphic.getProperties();
      p.shadow = store.get( DGraphic.P_3D_SHADOW, DGraphic.DEF_3D_SHADOW );
      p.light = store.get( DGraphic.P_3D_LIGHT, DGraphic.DEF_3D_LIGHT );
      //get contour line
      Object obj = store.get( DGraphic.P_XYS_CONTOUR_LINE, null );
      if( obj instanceof String )
      {
        if( p.z_axis!=null && DAxis.P_STEP_LINE.equals( obj ) ) 
        {
          p.contour_line = p.z_axis.getProperties().get( DAxis.P_STEP_LINE, -1f );
          if( p.contour_line > 0 )
          {
            //define delta in 0..1 space
            p.contour_line /= (p.surface.getZMax()-p.surface.getZMin());
          }
        }
      }
      else if ( obj instanceof IDColor )
      {
        IDColor clr = (IDColor)obj;
        p.contour_line = RGBA.Get( clr.getRed(), clr.getGreen(), clr.getBlue() );
      }
      else if ( obj instanceof Number )
      {
        p.contour_line = ((Number)obj).doubleValue();
        //delta is defined in z axis coordinate if this axis exists
        if( p.z_axis!=null && p.contour_line > 0 )
        {
          //define delta in 0..1 space
          p.contour_line /= (p.surface.getZMax()-p.surface.getZMin());
        }
      } else {
        //no contour line
        p.contour_line = -1f;
      }
      //get contour line color (colorizer)
      obj = store.get( DGraphic.P_XYS_CONTOUR_LINE_COLOR, null);
      if( obj instanceof IXYSurfaceColorizer ) {
        p.contour_line_colorizer = (IXYSurfaceColorizer)obj;
      } else {
        p.contour_line_colorizer = null;
        if( obj instanceof Integer ) {
          p.contour_line_rgba = ((Integer)obj).intValue();
        } else {
          p.contour_line_rgba = RGBA.BLACK; //default
        }
      }
      p.draw_polygon = !store.get( DGraphic.P_XYS_NO_POLYGON, false );
    }

    double h = rpd.graphic.getProperties().get( DGraphic.P_3D_HEIGHT, 1.0 );
    p.bbox = new Box3D( 0,1, 0,1, 0,h );

    //two angle to drive view
    double phi   = rpd.graphic.getProperties().get( DGraphic.P_XYZ_PHI,   0.3 );
    double theta = rpd.graphic.getProperties().get( DGraphic.P_XYZ_THETA, 0.2 );

    View3D view = new View3D( phi, theta, 0, 1.0f, p.bbox );
    //will be rescaled after insets computing
    view.scaleTo( rpd.ag_rect.x(), rpd.ag_rect.y(), rpd.ag_rect.w(), rpd.ag_rect.h() );
    
    // not enough values to render a one face surface ?
    if( p.surface.getSizeX() < 2 || p.surface.getSizeY()<2 ) return ;

    IScale z_scale = new LinearNumberScale(0, h, new Double(p.surface.getZMin()),new Double(p.surface.getZMax()) );
    z_scale.setLocale(rpd.locale);
    if( !z_scale.isValid() )
    {
      if( d.drawing() ) 
      {
        d.error("Invalid scale can't draw surface", rpd );
      }
      return ;
    }
    
    double ntheta = Radian.normalize( view.getTheta() );
    boolean x_incr = (ntheta <= Radian._PI2 || ntheta >= Radian._3PI2 );
    boolean y_incr = (0<= ntheta && ntheta<=Radian._PI);
    double nphi = Radian.normalize( view.getPhi() );
    boolean z_incr = (0<= nphi && nphi <= Radian._PI );
    if( nphi > Radian._PI2 && nphi < Radian._3PI2 )
    {
      x_incr =!x_incr;
      y_incr =!y_incr;
    }
    
    p.projected_contour_line = rpd.graphic.getProperties().get( DGraphic.P_XYZ_ZPROJ, false );
    p.zproj_contour_line = z_incr ? h : 0;
    
    //scale 3D view for rendering if gR rectange, this first scale to whole graphic rectangle
    //is mandatory to estimate axis insets in order to reduce 3D view rectangle and render axis.
    view.scaleTo( rpd.ag_rect.x(), rpd.ag_rect.y(), rpd.ag_rect.w(), rpd.ag_rect.h() );
    Insets insets = new Insets();
    if( p.z_axis!=null )
    {
      DefaultChartRender3D.renderZAxis3D( d, view, p.bbox, p.z_axis, z_scale, p.z_axis_style, x_incr, y_incr, z_incr, false, insets );
    }
    IScale x_scale=null;
    if( p.x_axis!=null )
    {
      Object xmin = p.x_axis.getProperties().get( DAxis.P_MIN, new Double(0) );
      Object xmax = p.x_axis.getProperties().get( DAxis.P_MAX, new Double(1) );
      x_scale = DefaultChartRenderData.getScale( p.x_axis, 0,1, xmin, xmax );
      if( !x_scale.isValid() )
      {
        if( d.drawing() ) 
        {
          d.error("Invalid X scale (missed DAxis.P_MIN and P_MAX properties ?)", rpd );
        }
        return ;
      }
      DefaultChartRender3D.renderXYAxis3D( true, d, view, p.bbox, p.x_axis, x_scale, p.x_axis_style, x_incr, y_incr, z_incr, false, insets );
    }
    IScale y_scale=null;
    if( p.y_axis!=null )
    {
      Object ymin = p.y_axis.getProperties().get( DAxis.P_MIN, new Double(0) );
      Object ymax = p.y_axis.getProperties().get( DAxis.P_MAX, new Double(1) );
      y_scale = DefaultChartRenderData.getScale( p.y_axis, 0,1, ymin, ymax );
      if( !y_scale.isValid() )
      {
        if( d.drawing() ) 
        {
          d.error("Invalid Y scale (missed DAxis.P_MIN and P_MAX properties ?)", rpd );
        }
        return ;
      }
      DefaultChartRender3D.renderXYAxis3D( false, d, view, p.bbox, p.y_axis, y_scale, p.y_axis_style, x_incr, y_incr, z_incr, false, insets );
    }
    view.scaleTo( rpd.ag_rect.x()+insets.getL(), rpd.ag_rect.y()+insets.getT(),  
                  rpd.ag_rect.w()-insets.getW(), rpd.ag_rect.h()-insets.getH() );

    Vector3D view_vector = view.getViewVector(null);
    view_vector.normalize();

    if( d.drawing() )
    {
      //background of bar bounding box
      DefaultChartRender3D.drawBox3D( d, view, p.bbox, x_incr, y_incr, z_incr, false );
    }

    //background part of axis     
    if( p.x_axis!=null )
      DefaultChartRender3D.renderXYAxis3D( true, d, view, p.bbox, p.x_axis, x_scale, p.x_axis_style, x_incr, y_incr, z_incr, false, null );    
    if( p.y_axis!=null )
      DefaultChartRender3D.renderXYAxis3D( false, d, view, p.bbox, p.y_axis, y_scale, p.y_axis_style, x_incr, y_incr, z_incr, false, null );    
    if( p.z_axis!=null )
      DefaultChartRender3D.renderZAxis3D( d, view, p.bbox, p.z_axis, z_scale, p.z_axis_style, x_incr, y_incr, z_incr, false, null );
    
    renderZMap(  z_scale, d,rpd,p, x_incr,y_incr,z_incr, view, view_vector, h );
    
    if( d.drawing() )
    {
      //front of bars bounding box
      DefaultChartRender3D.drawBox3D( d, view, p.bbox, x_incr, y_incr, z_incr, true );
    }
    //forground part of axis (and location)
    if( p.x_axis !=null )
      DefaultChartRender3D.renderXYAxis3D( true, d, view, p.bbox, p.x_axis, x_scale, p.x_axis_style, x_incr, y_incr, z_incr, true, null );    
    if( p.y_axis!=null )
      DefaultChartRender3D.renderXYAxis3D( false, d, view, p.bbox, p.y_axis, y_scale, p.y_axis_style, x_incr, y_incr, z_incr, true, null );    
    if( p.z_axis !=null )
      DefaultChartRender3D.renderZAxis3D( d, view, p.bbox, p.z_axis, z_scale, p.z_axis_style, x_incr, y_incr, z_incr, true, null );    
  }
  
  private static H3D rebuildPersistData( DefaultChartRenderData d, RenderPersistData rpd )
  {
      H3D p = new H3D();
      rpd.g = p;

      //retrieve the 3 required axis ...
      p.z_axis = null;
      p.surface=null;
      int missing=4;
      //get axis and IXYSurface 
      for( IDItem item = rpd.graphic.getFirstChild(); item!=null; item=item.getNext() )
      {
        if( item instanceof DAxis )
        {
          String name = ((DAxis)item).getName();
          if( "x".equals( name ) && p.x_axis == null ) { p.x_axis=(DAxis)item; missing--; }
          if( "y".equals( name ) && p.y_axis == null ) { p.y_axis=(DAxis)item; missing--; }
          if( "z".equals( name ) && p.z_axis == null ) { p.z_axis=(DAxis)item; missing--; }
        }
        else if ( item instanceof IXYSurface )
        {
          if( p.surface==null ) { p.surface=(IXYSurface)item; missing--; }
        }
        if( missing==0 ) break;
      }

      if( p.x_axis !=null ) p.x_axis_style = new IGCDStyle( p.x_axis );
      if( p.y_axis !=null ) p.y_axis_style = new IGCDStyle( p.y_axis );
      if( p.z_axis !=null ) p.z_axis_style = new IGCDStyle( p.z_axis );
      
      if( p.surface instanceof IDItem )
      {
        p.loc_item = (IDItem)p.surface;
      } else {
        p.loc_item = rpd.graphic;//by default
      }
      
      return  p;
  }

  private static IXYSurfaceColorizer default_colorizer_;
  private static IXYSurfaceColorizer getDefaultColorizer()
  {
    if( default_colorizer_ == null )
    {
      default_colorizer_ =  new IXYSurfaceColorizer() 
      {
         GradientGenerator gg = new GradientGenerator( new int[]{RGBA.RED,RGBA.YELLOW, RGBA.GREEN });
         public int getRGBA(double x, double y, double z, double cos_angle ) 
         {
           return gg.getRGBA( z );
         }
         public boolean isApplyShadow() { return true; }
      };
    }
    return default_colorizer_;
  }

  //polygon using 3 points references
  //use Polygon as it's faster for SWTGC users, coordinates are synchronized on update() call
  private static class PrivPolygon extends Polygon
  {
     Point2D p0, p1, p2; //ref!     
     public PrivPolygon( Point2D pt0, Point2D pt1, Point2D pt2 )
     {
       super(4);
       p0=pt0; p1=pt1; p2=pt2;
     }
    public int getPolySize() { return 3; }
    public int getPolyX(int index) {
      switch( index ) {
      case 0: return (int)p0.getX();
      case 1: return (int)p1.getX();
      case 2: return (int)p2.getX();
      case 4: return (int)p0.getX();
      }
      return 0;
    }
    public int getPolyY(int index) {
      switch( index ) {
      case 0: return (int)p0.getY();
      case 1: return (int)p1.getY();
      case 2: return (int)p2.getY();
      case 4: return (int)p0.getY();
      }
      return 0;
    }
    public boolean isPolyClosed() { return true; }
    public IShape copyShape() { return null; }

    public void update()
    {
      points_xy_[0] = (int)p0.getX();   points_xy_[1] = (int)p0.getY();
      points_xy_[2] = (int)p1.getX();   points_xy_[3] = (int)p1.getY();
      points_xy_[4] = (int)p2.getX();   points_xy_[5] = (int)p2.getY();
      points_xy_[6] = points_xy_[0];    points_xy_[7] = points_xy_[1];        
    }

    public IRect getBounds() {
//TODO... if needed
      return null;
    }
    
  }
  private static void renderZMap( IScale scale, DefaultChartRenderData d, RenderPersistData rpd, H3D p, boolean x_incr, boolean y_incr, boolean z_incr, View3D view, Vector3D view_vector, double H )
    throws DefaultRenderChartLocation
  {
    IXYSurface surf = p.surface;
    final int size_x = surf.getSizeX();
    final int size_y = surf.getSizeY();
    int icrv0, icrvX, icrvI,icrvL;
    int icrv_last;
    if( y_incr ) {
      icrv_last=0;
      icrv0 = 1 ;
      icrvX = size_y;
      icrvI = +1;
      icrvL = icrvX-1;
    } else {
      icrv_last = size_y-1;
      icrv0 = icrv_last-1;
      icrvX = -1;
      icrvI = -1;
      icrvL = 0;
    }
    int lines_rgba = RGBA.BLACK;
    if( d.drawing() ) 
    {
//TODO: replace BLACK from properties... (?P_LINE_COLOR)
//use X axis P_LINE_COLOR and Y axis P_LINE_COLOR if any
//use Z axis P_LINE_COLOR
//use BLACK as default    
      if( p.contour_line<=0 ) {
        d.gc_.setPen( new LineStylePen( lines_rgba ) );
      }
    }

    Point2D p00 = new Point2D(); //pXY, 0:last, 1:current
    Point2D p01 = new Point2D();
    Point2D p10 = new Point2D();
    Point2D p11 = new Point2D();
    int  poly_type;
    PrivPolygon p1, p2;
    if( x_incr ) 
    {
      if( y_incr )
      {
        poly_type=1;
        p1 = new PrivPolygon( p00, p10, p01 );
        p2 = new PrivPolygon( p01, p10, p11 );
      } else {
        poly_type=2;
        p1 = new PrivPolygon( p01, p11, p00 );
        p2 = new PrivPolygon( p00, p11, p10 );
      }
    } else {
      if( y_incr ) {
        poly_type=3;
        p1 = new PrivPolygon( p10, p00, p11 );
        p2 = new PrivPolygon( p11, p00, p01 );
      } else {
        poly_type=4;
        p1 = new PrivPolygon( p11, p01, p10 );
        p2 = new PrivPolygon( p10, p01, p00 );
      }
    }
    Object prop = rpd.graphic.getProperties().get( DGraphic.P_XYS_COLORIZER, null );
    IXYSurfaceColorizer clr=null;
    if( prop instanceof IXYSurfaceColorizer )
    {
      clr = (IXYSurfaceColorizer) prop;
    } else {
      clr = getDefaultColorizer();
    }
    
    Vector3D normal = new Vector3D();
    int lines_x = rpd.graphic.getProperties().get(DGraphic.P_XYS_LINE_X,0);
    int lines_y = rpd.graphic.getProperties().get(DGraphic.P_XYS_LINE_Y,0);
    boolean lines_enabled = d.drawing() && (lines_x>0 || lines_y>0);
    boolean apply_shadow = clr.isApplyShadow();
LineStylePen pen = new LineStylePen();
    VDouble vcz=new VDouble();
    VDouble vlz=new VDouble();
    boolean do_line_x=false, do_line_y=false;
    for( int icrv=icrv0 ; icrv!=icrvX; icrv_last=icrv, icrv += icrvI )
    {
      double ly = icrv_last / (double)(size_y-1);
      double cy = icrv      / (double)(size_y-1);
      
      int ipt0, iptX, iptI,iptL ;
      if( x_incr ) {
        ipt0 = 0;
        iptX = size_x;
        iptI = +1;
        iptL = size_x-1;
      } else {
        ipt0 = size_x-1;
        iptX = -1;
        iptI = -1;
        iptL = 0;
      }
      if( lines_y>0 ) {
        do_line_y = (icrv % lines_y)==0;
      }
      double lx=0, lzly=0, lzcy=0;
      for( int ipt=ipt0; ipt!=iptX; ipt+=iptI )
      {
        double cx = ipt      / (double)(size_x-1);
        vlz.setValue( surf.getZ( ipt, icrv_last ) );
        vcz.setValue( surf.getZ( ipt, icrv     ) );
        double czly = scale.toScale( vlz );
        double czcy = scale.toScale( vcz );
        view.projection( cx, ly, czly , p10 );
        view.projection( cx, cy, czcy,  p11 );  
        if( lines_x>0 ) {
          do_line_x = (ipt%lines_x)==0;
        }
        if( ipt!=ipt0 )
        {
          int rgba=0;
          double cos_angle=0;
          if( d.drawing() )
          {
            double ctrx = (lx+cx)/2;
            double ctry = (ly+cy)/2;            
            double ctrz = (lzly+czly+lzcy+czcy)/4;
            normal.vectorialProduct( 0,cy-ly,lzcy-lzly,  cx-lx,0,czcy-lzcy );
            normal.normalize();
            
            cos_angle = Vector3D.Scalar( normal, view_vector );
            rgba = clr.getRGBA( ctrx,ctry,ctrz/H, cos_angle );
            if( apply_shadow ) 
              rgba=RGBA.Shadow( rgba, p.shadow, p.light, (float)Math.abs(cos_angle) );
          }
          
          //can render 2 faces for this "quad"
          switch( poly_type )
          {
          case 1:
            //p1:p00, p10, p01
            renderPolygon( d,p,p1, lx,ly,lzly, cx,ly,czly, lx,cy,lzcy, normal,view_vector,view, H, apply_shadow,rgba,cos_angle );
            //p2:  p01, p10, p11
            renderPolygon( d,p,p2, lx,cy,lzcy, cx,ly,czly, cx,cy,czcy, normal,view_vector,view, H, apply_shadow,rgba,cos_angle );
          break;
          case 2:
            //p1: p01, p11, p00 
            renderPolygon( d,p,p1, lx,cy,lzcy, cx,cy,czcy, lx,ly,lzly, normal,view_vector,view, H, apply_shadow,rgba,cos_angle );
            //p2:  p00, p11, p10 
            renderPolygon( d,p,p2, lx,ly,lzly, cx,cy,czcy, cx,ly,czly, normal,view_vector,view, H, apply_shadow,rgba,cos_angle );
          break;
          case 3:
            //p1: p10, p00, p11
            renderPolygon( d,p,p1, cx,ly,czly, lx,ly,lzly, cx,cy,czcy, normal,view_vector,view, H, apply_shadow,rgba,cos_angle );
            //p2: p11, p00, p01
            renderPolygon( d,p,p2,  cx,cy,czcy, lx,ly,lzly, lx,cy,lzcy,  normal,view_vector,view, H, apply_shadow,rgba,cos_angle );
          break;
          case 4:
            //p1: p11, p01, p10
            renderPolygon( d,p,p1, cx,cy,czcy, lx,cy, lzcy, cx,ly,czly, normal,view_vector,view, H, apply_shadow,rgba,cos_angle );
            //p2: p10, p01, p00
            renderPolygon( d,p,p2, cx,ly,czly, lx,cy, lzcy, lx,ly,lzly, normal,view_vector,view, H, apply_shadow,rgba,cos_angle );
          }
          
          if( lines_enabled ) {
            if( p.contour_line>0 ) {
              d.gc_.setPen( new LineStylePen( lines_rgba ) );
            }
            if(do_line_x)d.gc_.drawLine( (int)p00.getX(),(int)p00.getY(), (int)p01.getX(),(int)p01.getY() );
            if(do_line_y)d.gc_.drawLine( (int)p00.getX(),(int)p00.getY(), (int)p10.getX(),(int)p10.getY() );            
            if( ipt==iptL ) {
              if(do_line_y)d.gc_.drawLine( (int)p01.getX(),(int)p01.getY(), (int)p11.getX(),(int)p11.getY() );
            }
            if( icrv==icrvL ) {
              if(do_line_x)d.gc_.drawLine( (int)p10.getX(),(int)p10.getY(), (int)p11.getX(),(int)p11.getY() );
            }
          }
        }
        
        p00.setPoint( p10 );
        p01.setPoint( p11 );
        lx=cx;
        lzcy=czcy;
        lzly=czly;
      }
    }
  }
  private static boolean haveLevelZ( double lz, double x1,double y1,double z1, double x2, double y2,double z2, Point3D res )
  {
    if( z1==z2 ) return false; //even lz==z1 as I don't want to choose x/y (all valid)
    double zn,zx;
    if( z1 <= z2 ) {
      zn = z1; zx=z2;
    } else {
      zn = z2; zx=z1;
    }
    if( lz< zn || lz > zx ) return false;
    double k = (lz-z1)/(z2-z1);
    //if( Double.isNaN(k) ) return false;
    double x = x1+k*(x2-x1);
    double y = y1+k*(y2-y1);
    res.setCoord( x, y, lz );
    return true;
  }
  private static SolidBrush brush ;
  private static Point2D rpp0 = new Point2D(), rpp1 = new Point2D();
  private static Point3D rp0 = new Point3D(), rp1 = new Point3D();
  private static void renderPolygon( DefaultChartRenderData d, H3D p, PrivPolygon poly, double x1, double y1, double z1, double x2, double y2, double z2, double x3, double y3, double z3, Vector3D normal, Vector3D view_vector, View3D view, double H, boolean apply_shadow, int rgba, double cos_angle )
  throws DefaultRenderChartLocation
  {
    poly.update();
    if( d.drawing() ) 
    {
      if( p.draw_polygon )
      {
        if( brush==null ) brush = new SolidBrush();
        brush.setRGBA(rgba);      
        d.gc_.setBrush(brush);
        d.gc_.fillPoly( poly );
      }
      //draw contour lines
      if( p.contour_line > 0 )
      {
        if( p.contour_line_colorizer==null ) 
        {
          pen.setRGBA( p.contour_line_rgba );
          d.gc_.setPen( pen );
        }
        double zn=z1, zx=z1;
        if( z2 < zn ) zn = z2; else if ( z2 > zx) zx=z2;
        if( z3 < zn ) zn = z3; else if ( z3 > zx) zx=z3;

//TODO: get first lz >= zn in a smarter way please ...        
        //double lz = DZ*((int)(z3/DZ));
        double lz=0.0;
        while( lz < zn ) lz+= p.contour_line;
        while( lz<=zx )
        {
          boolean both=false;
          if( haveLevelZ( lz, x1,y1,z1, x2,y2,z2, rp0 ))
          {
            if( haveLevelZ( lz, x1,y1,z1, x3,y3,z3, rp1)
                ||haveLevelZ( lz, x2,y2,z2, x3,y3,z3, rp1 ))
            {
              both=true;
            }
          }
          else if( haveLevelZ( lz, x2,y2,z2, x3,y3,z3, rp0 )
                && haveLevelZ( lz, x1,y1,z1, x3,y3,z3, rp1 ))
          {
            both=true;
          }
          if( both )
          {
            if( p.contour_line_colorizer!=null ) 
            {
              pen.setRGBA( p.contour_line_colorizer.getRGBA( rp0.getX(), rp0.getY(),lz, cos_angle));
              d.gc_.setPen( pen );
            }
            if( p.projected_contour_line )
            {
              view.projection( rp0.getX(),rp0.getY(),p.zproj_contour_line, rpp0 );
              view.projection( rp1.getX(),rp1.getY(),p.zproj_contour_line, rpp1 );
            } else {
              view.projection( rp0.getX(),rp0.getY(),rp0.getZ(), rpp0 );
              view.projection( rp1.getX(),rp1.getY(),rp1.getZ(), rpp1 );
            }
            d.gc_.drawLine( (int)rpp0.getX(),(int)rpp0.getY(),  (int)rpp1.getX(),(int)rpp1.getY() );
          }
          lz += p.contour_line;
        }
      } 
    }
    else //locating
    {
       if( poly.contains( d.lx_, d.ly_ ) ) {
         String loc_id  =DLocated.CurvePoint;
         throw new DefaultRenderChartLocation( loc_id, p.loc_item, new Rect(d.lx_,d.ly_,0,0) );
       }   
     }
  }

//TODO:remove debug method here:  
  private static LineStylePen pen = new LineStylePen();
  private static void drawNormalVector( IGC gc, Vector3D normal, double sx,double sy,double sz, View3D view, int rgba )
  {
    Point2D pt1 = new Point2D(), pt2=new Point2D();
    view.projection( sx,sy,sz, pt1);
    normal.normalize();
    float k=0.2f;
    view.projection( sx+k*normal.getX(),sy+k*normal.getY(),sz+k*normal.getZ(), pt2);
    pen.setRGBA( rgba );
    IPen ppp = gc.setPen( pen );
    gc.drawLine( (int)pt1.getX(),(int)pt1.getY(), (int)pt2.getX(),(int)pt2.getY() );
    gc.setPen(ppp);
  }
  
}
