/*******************************************************************************
 * Copyright (c) 2005, 2011 - IBM, Intel Corporation.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 *
 * Contributors:
 *    Samson Wai, IBM - Initial implementation
 *    Andy Kaylor, Intel - Refactored code for shared library
 *
 * $Id: TPTPLogFile.cpp,v 1.1 2005-09-12 17:52:55 akaylor Exp $ 
 ******************************************************************************/ 

#include <string.h>
#include <stdlib.h>

#include "tptp/TPTPLogUtils.h"
#include "TPTPLogUtilsInternal.h"
#include "tptp/TPTPUtils.h"

tptp_int32 tptp_createLogFile( tptp_string* filename, tptp_filehandle* f )
{
	*f = fopen(filename, "w");
	if (*f == NULL) 
	{
		return TPTP_ERROR_CANNOT_OPEN_LOG_FILE;
	}

	return 0;
}

/* Bug 306136 - Implement log rollover */
tptp_filehandle tptp_rolloverLogFile( tptp_filehandle f, tptp_int32 history, tptp_int32 size, char* baseFile )
{
	char** filenames = NULL;
	int i;
	
	if( (tptp_filehandle)-1 == f || NULL == f )
	{
		return f;
	}
	
	if( ftell(f) < size ) 
	{
		return f;
	}
	fclose(f);
	
	/* build list of filenames */
	filenames = (char**)tptp_malloc( sizeof(char*) * (history+1) );
	
	/* If we failed to allocate memory then we will fall through and at least
	   truncate the existing log file; this prevents the log file from growing 
	   insanely large and filling up the disk. */
	if( filenames ) 
	{
		filenames[0] = baseFile;
		for( i=1; i<history+1; i++ ) 
		{
			
			char n[16]; /* Enough space for an absurdly large history (10^14 - 1) */
			char* spliceAt = strrchr(baseFile, '.');
			int spliceIndex;
			
			if( !spliceAt )
				spliceAt = &baseFile[strlen(baseFile)+1];
			spliceIndex = spliceAt - baseFile;
			
			sprintf(n, "_%d", i);
			filenames[i] = (char*)tptp_malloc( strlen(baseFile) + strlen(n) + 1 );
			
			/* Its not going to behave well if we run out of memory; but at least we won't crash */
			if( filenames[i] )
			{
				memcpy(&filenames[i][0], baseFile, spliceAt - baseFile); /* copy from baseFile[0 to '.'] */
				memcpy(&filenames[i][spliceIndex], n, strlen(n)); 	/* append the _%n counter */
				memcpy(&filenames[i][spliceIndex+strlen(n)], spliceAt, strlen(baseFile) - spliceIndex + 1); /* append extension + null terminator */	
			}
		}
		
		/* Cycle the history files; if any failures occurred allocating memory, then
		   some of these operations will be skipped. It won't result in proper rollover,
		   but at least we won't crash. */
		if( filenames[history] ) /* Remove oldest in history */
			remove(filenames[history]);
		for( i=history; i>0; i-- ) /* Mark the remaining files as one older */
		{
			if( filenames[i-1] && filenames[i] )
			{
				rename(filenames[i-1], filenames[i]);
			}
		}
		
		/* Free up the file names */
		for( i=1; i<history+1; i++ ) 
		{
			if( filenames[i] )
				tptp_free(filenames[i]);
		}
		tptp_free(filenames);
	}
	
	/* Open a new instance of the log file */
	f = fopen(baseFile, "w");
	return f;
}

tptp_int32 tptp_writeToLogFile( tptp_filehandle f, const tptp_string* cbe, tptp_int32* bytesWritten )
{
	if (f == NULL) 
	{
		return TPTP_ERROR_INVALID_FILE_HANDLE;
	}


	*bytesWritten = fprintf( f, "%s\n", cbe );
	fflush( f );

	if ( *bytesWritten < 0 )
	{
		return TPTP_ERROR_WRITING_TO_FILE;
	}

	return 0;
}

tptp_int32 tptp_closeLogFile( tptp_filehandle f )
{
	int rc = fclose( f );

	if ( rc == 0 ) 
	{
		return TPTP_ERROR_CLOSING_LOG_FILE;
	}

	return 0;
}


